//#define _FILE_OFFSET_BITS 64

#include <sys/ptrace.h>
#include <signal.h>
#include <stdio.h>
#include <fcntl.h>
#include <stdarg.h>
#include <asm/ldt.h>
#include <asm/ucontext.h>
#include <linux/unistd.h>
#include <linux/elf.h>

#include "cryopid.h"
#include "cpimage.h"
#include "process.h"

static int image_fd;

int verbosity = 0;
int do_pause = 0;
int want_pid = 0;
int translate_pids = 0;
int tls_hack = 0;
long tls_start = 0;
char ignorefds[256];
int reforked = 0;

int real_argc;
char** real_argv;
char** real_environ;
extern char** environ;

void safe_read(int fd, void* dest, size_t count, char* desc) {
    int ret;
    ret = read(fd, dest, count);
    if (ret == -1) {
	fprintf(stderr, "Read error on %s: %s\n", desc, strerror(errno));
	exit(1);
    }
    if (ret < count) {
	fprintf(stderr, "Short read on %s\n", desc);
	exit(1);
    }
}

#if !set_thread_area
_syscall1(int,set_thread_area,struct user_desc*,u_info);
#endif

int set_rt_sigaction(int sig, const struct k_sigaction* ksa, const struct k_sigaction* oksa) {
    int ret;
    asm (
	    "int $0x80"
	    : "=a"(ret)
	    : "a"(__NR_rt_sigaction), "b"(sig),
	      "c"(ksa), "d"(oksa), "S"(sizeof(ksa->sa_mask))
	);
    return ret;
}

int voodoo_stage = 0;
char* voodoo_ptr = NULL;
char voodoo_backup[12];
void segv_handler(int sig, siginfo_t *si, void *ucontext) {
    struct ucontext *uc = (struct ucontext*)ucontext;
    unsigned char *pt = (unsigned char*)uc->uc_mcontext.eip;
    if (voodoo_stage == 1) {
	pt = voodoo_ptr;
	pt[0] = 0xa3;
	*(long*)(pt+1) = 0x00000000;
	pt[5] = 0x90;
	pt[6] = 0xeb;
	pt[7] = 0xf8;
	voodoo_stage++;
	return;
    } else if (voodoo_stage == 2) {
	*(long*)pt = 0x90909090;
	pt[4] = 0xa3;
	*(long*)(pt+5) = 0x00000000;
	voodoo_stage++;
	return;
    } else if (voodoo_stage == 3) {
	memcpy(voodoo_ptr, voodoo_backup, sizeof(voodoo_backup));
	voodoo_stage = 0;
	return;
    }
    if (!memcmp(pt, "\x65\x83\x3d", 3)) {
	/*
	 *  8048344:   65 83 3d 0c 00 00 00    cmpl   $0x0,%gs:0xc
	 *  804834b:   00 
	 *  804834c:   83 3d af be ad de 00    cmpl   $0x0,0xdeadbeaf
	 */
	pt[0] = 0x83;
	pt[1] = 0x3d;
	*(long*)(pt+2) = tls_start+*(char*)(pt+3);
	pt[6] = pt[7];
	pt[7] = 0x90;
	return;
    }
    if (!memcmp(pt, "\x65\x8b", 2) && (
	    pt[2] == 0x0d || /* ecx */
	    pt[2] == 0x35 || /* esi */
	    pt[2] == 0x15 || /* edx */
	    pt[2] == 0x2d || /* ebp */
	    pt[2] == 0x3d || /* edi */
	    0)) {
	/*
	 *  8048353:   65 8b 0d 00 00 00 00    mov    %gs:0x0,%ecx
	 *  804835a:   8b 0d af be ad de       mov    0xdeadbeaf,%ecx
	 */
	pt[0] = 0x8b;
	pt[1] = pt[2];
	*(long*)(pt+2) = tls_start+*(long*)(pt+3);
	pt[6] = 0x90;
	return;
    }
    if (!memcmp(pt, "\x65\x89\x3d", 3)) { /* XXX untested */
	/*
	 * 80483ab:   65 89 3d 50 00 00 00    mov    %edi,%gs:0x50
	 * 80483b2:   89 3d ef be ad de       mov    %edi,0xdeadbeef
	 */
	pt[0] = 0x89;
	pt[1] = 0x3d;
	*(long*)(pt+2) = tls_start+*(long*)(pt+3);
	pt[6] = 0x90;
	return;
    }
    if (!memcmp(pt, "\x65\xc7\x05", 3)) { /* XXX untested */
	/*
	 *  8048344:   65 c7 05 f0 01 00 00    movl   $0xffffffff,%gs:0x1f0
	 *  804834b:   ff ff ff ff
	 *  804834f:   c7 05 ef be ad de ff    movl   $0xfffffff,0xdeadbeef
	 *  8048356:   ff ff 0f
	 */
	pt[0] = pt[1];
	pt[1] = pt[2];
	*(long*)(pt+2) = tls_start+*(long*)(pt+3);
	*(long*)(pt+6) = *(long*)(pt+7);
	pt[10] = 0x90;
	return;
    }
    if (!memcmp(pt, "\xf0\x65\x0f\xb1\x0d", 5)) { /* XXX untested */
	/*
	 * 8048359:   f0 65 0f b1 0d 54 00    lock cmpxchg %ecx,%gs:0x54
	 * 8048360:   00 00
	 * 8048362:   f0 0f b1 0d ef be ad    lock cmpxchg %ecx,0xdeadbeef
	 * 8048369:   de
	 */
	pt[1] = 0x0f;
	pt[2] = pt[3];
	pt[3] = pt[4];
	*(long*)(pt+4) = tls_start+*(long*)(pt+5);
	pt[8] = 0x90;
	return;
    }
    if (!memcmp(pt, "\xf0\x65\x83\x0d", 4)) { /* XXX untested */
	/*
	 * 804836a:   f0 65 83 0d 54 00 00    lock orl $0x10,%gs:0x54
	 * 8048371:   00 10
	 * 8048373:   f0 83 0d ef be ad de    lock orl $0x10,0xdeadbeef
	 * 804837a:   10
	 */
	pt[1] = pt[2];
	pt[2] = pt[3];
	*(long*)(pt+3) = tls_start+*(long*)(pt+4);
	pt[7] = pt[8];
	pt[8] = 0x90;
	return;
    }
    if (pt[0] == 0x65 && (
		pt[1] == 0xa1 || /* mov    %gs:0x0,%eax  */
		pt[1] == 0xa3    /* mov    %eax,%gs:0x48 */
		)) {
	/*
	 * 804838c:   65 a1 00 00 00 00       mov    %gs:0x0,%eax
	 * 8048392:   a1 ef be ad de          mov    0xdeadbeef,%eax
	 *
	 * 80483c8:   65 a3 48 00 00 00       mov    %eax,%gs:0x48
	 * 80483ce:   a3 ef be ad de          mov    %eax,0xdeadbeef
	 */
	pt[0] = pt[1];
	*(long*)(pt+1) = tls_start+*(long*)(pt+2);
	pt[5] = 0x90;
	return;
    }
    if (!memcmp(pt, "\x65\x89\x51", 3)) {
	/*
	 * 80483b1:   65 89 51 00             mov    %edx,%gs:0x0(%ecx)
	 * 80483b5:   89 91 af be ad de       mov    %edx,0xdeadbeaf(%ecx)
	 *
	 * WARNING: XXX VOODOO HAPPENS HERE
	 */
	memcpy(voodoo_backup, pt, sizeof(voodoo_backup));
	pt[0] = 0x89;
	pt[1] = 0x91;
	*(long*)(pt+2) = tls_start+*(char*)(pt+3);
	pt[6] = 0xa3;
	*(long*)(pt+7) = 0x00000000; /* cause another segfault */
	voodoo_stage = 1;
	voodoo_ptr = pt;
	return;
    }
    if (old_segvhandler &&
	    old_segvhandler != (void*)SIG_IGN && old_segvhandler != (void*)SIG_DFL)
	old_segvhandler(sig, si, ucontext);
    printf("Unhandled segfault at 0x%08lx!\n", uc->uc_mcontext.eip);
    _exit(1);
}

void resume_image_from_file(int fd) {
    int num_maps, num_fds, num_tls;
    int fd2;
    int stdinfd;
    struct user_desc tls;
    struct map_entry_t map;
    struct fd_entry_t fd_entry;
    pid_t pid;
    struct user user_data;
    struct user_i387_struct i387_data;
    struct k_sigaction sa;
    char dir[1024];
    int cmdline_length, environ_length;
    char *new_cmdline, *new_environ;
    int i;
    int extra_prot = 0;

    safe_read(fd, &pid, sizeof(pid), "pid");

    extern int start_supervisor(pid_t);
    if (translate_pids)
	start_supervisor(pid);

    if (want_pid) {
	if (kill(pid, 0) == 0 || errno == EPERM) {
	    fprintf(stderr, "Pid is already taken. Refusing to do anything :(\n");
	    exit(1);
	}
	/* grab it! */
	if (verbosity > 0)
	    fprintf(stderr, "PID is free ... going for it!\n");
	int happy = 0;
	switch(fork()) {
	    case 0: break;
	    case -1: perror("fork"); exit(1);
	    default:
		     {
			 sigset_t mask;
			//printf("myppid() = %d\n", getpid());
			 sigfillset(&mask);
			 sigdelset(&mask, SIGUSR1);
			 sigsuspend(&mask);
			 wait(NULL);
			 _exit(0);
		     }
	}
	while (!happy) {
	    switch(fork()) {
		case 0:
		    setpgid(getpid(), getppid());
		    if (getpid() == pid) happy = 1;
		    break;
		case -1:
		    perror("fork");
		    exit(1);
		default:
		    sleep(1);
		    _exit(0);
	    }
	}
	//printf("getppid() = %d\n", getppid());
	kill(getppid(),SIGUSR1);
    }

    new_cmdline = (char*)malloc(65536);
    new_environ = (char*)malloc(65536);
    safe_read(fd, &cmdline_length, sizeof(cmdline_length), "cmdline_length");
    safe_read(fd, new_cmdline, cmdline_length, "cmdline");
    new_cmdline[cmdline_length]='\0';

    safe_read(fd, &environ_length, sizeof(environ_length), "environ_length");
    safe_read(fd, new_environ, environ_length, "environ");
    new_environ[environ_length]='\0';
    
    if (!reforked) {
	char fn[128];
	char *p;
	char **new_argv, **new_envp;
	new_argv = (char**)malloc(sizeof(char*)*4096); /* FIXME */
	new_envp = (char**)malloc(sizeof(char*)*4096); /* FIXME */

	p = new_cmdline;
	i = 0;
	while (*p) {
	    new_argv[i++] = p;
	    while(*p++);
	}
	new_argv[i] = NULL;

	p = new_environ;
	i = 0;
	while (*p) {
	    new_envp[i++] = p;
	    while(*p++);
	}
	new_envp[i] = NULL;

	snprintf(fn, 128, "/tmp/cryopid.state.%d.XXXXXX", getpid());
	int f = syscall_check(mkstemp(fn), 0, "mkstemp");
	unlink(fn);
	int i;
	syscall_check(write(f, &real_argc, sizeof(real_argc)), 0, "write");
	for(i=0; i < real_argc; i++) {
	    int l = strlen(real_argv[i])+1;
	    syscall_check(write(f, &l, sizeof(l)), 0, "write");
	    syscall_check(write(f, real_argv[i], l), 0, "write");
	}
	syscall_check(dup2(f, 42), 0, "dup2");
	syscall_check(fcntl(42, F_SETFD, 0), 0, "fcntl"); /* disable close-on-exec */
	if (verbosity > 0)
	    fprintf(stderr, "Re-exec'ing...\n");
	execve(real_argv[0], new_argv, new_envp);
	printf("execve(%s,0x%p,0x%p) failed. Not restoring cmdline. (error: %s)\n",
		real_argv[0], new_argv, new_envp, strerror(errno));
	close(f);
	close(42);
    }

    safe_read(fd, &user_data, sizeof(struct user), "user data");
    safe_read(fd, &i387_data, sizeof(struct user_i387_struct), "i387 data");
    safe_read(fd, &num_maps, sizeof(int), "num_maps");
    safe_read(fd, &num_tls, sizeof(int), "num_tls");

    if (num_tls > 0) {
	/* detect a TLS-capable system */
	if (set_thread_area(NULL) == -1) { /* you'd hope it did!*/
	    if (errno == ENOSYS) {
		tls_hack = 1;
		extra_prot = PROT_WRITE;
		if (verbosity > 0)
		    printf("TLS hack enabled!\n");
	    }
	}
    }

    if (verbosity > 0)
	fprintf(stderr, "Reading %d maps...\n", num_maps);

    while(num_maps--) {
	safe_read(fd, &map, sizeof(struct map_entry_t), "a map");

	if (map.filename && map.filename[0]) {
	    if (verbosity > 0)
		fprintf(stderr, "Loading 0x%08lx (len %ld) from file %s\n",
			map.start, map.length, map.filename);
	    syscall_check(fd2 = open(map.filename, O_RDONLY), 0,
		    "open(%s)", map.filename);
	    syscall_check((int)mmap((void*)map.start, map.length, map.prot|extra_prot,
		    MAP_FIXED|map.flags, fd2, map.pg_off),
		0, "mmap(0x%lx, 0x%lx, %x, %x, %ld, 0x%lx)",
		map.start, map.length, map.prot|extra_prot, map.flags, fd2, map.pg_off);
	    syscall_check(close(fd2), 0, "close(%d)", fd2);
	} else {
	    if (verbosity > 1)
		fprintf(stderr, "Creating 0x%08lx (len %ld) from image\n",
			map.start, map.length);
	    syscall_check( (int)
		mmap((void*)map.start, map.length, map.prot|extra_prot,
		    MAP_ANONYMOUS|MAP_FIXED|map.flags, -1, 0),
		    0, "mmap(0x%lx, 0x%lx, %x, %x, -1, 0)",
		    map.start, map.length, map.prot|extra_prot, map.flags);
	}
	if (map.data != NULL) {
	    if (verbosity > 0)
		fprintf(stderr, "Loading 0x%08lx (len %ld) from image\n",
			map.start, map.length);
	    safe_read(fd, (void*)map.start, map.length, "map data");
	}
    }

    if (verbosity > 0)
	fprintf(stderr, "Reading %d TLS entries...\n", num_tls);
    if (do_pause) sleep(1);
    for(i = 0; i < num_tls; i++) {
	safe_read(fd, &tls, sizeof(struct user_desc), "tls info");

	if (!tls.base_addr) continue;

	if (verbosity > 0)
	    fprintf(stderr, "Restoring TLS entry %d (0x%lx limit 0x%x)\n",
		    tls.entry_number, tls.base_addr, tls.limit);

	if (tls_hack)
	    tls_start = tls.base_addr;
	else
	    syscall_check(set_thread_area(&tls), 0, "set_thread_area");
    }

    stdinfd = 0; /* we'll use stdin for stdout/stderr later if needed */
    if (verbosity == 0) {
	close(1);
	close(2);
    }
    safe_read(fd, &num_fds, sizeof(int), "num_fds");
    if (verbosity > 0)
	fprintf(stderr, "Reading %d file descriptors...\n", num_fds);
    while (num_fds--) {
	safe_read(fd, &fd_entry, sizeof(struct fd_entry_t), "an fd_entry");
	if (verbosity > 0) {
	    fprintf(stderr, "fd: name=%s  fd=%d  mode=%d  flags=%d\n", fd_entry.filename, fd_entry.fd, fd_entry.mode, fd_entry.flags);
	}

	/* shuffle fd's about if they're going to clash with anything 
	 * important (stdin, or our image file)
	 */
	if (fd_entry.fd == fd) fd = dup(fd);
	if (fd_entry.fd == stdinfd) stdinfd = dup(stdinfd);

	if (fd_entry.flags & FD_IS_TERMINAL) {
	    if (verbosity > 1)
		fprintf(stderr, "    this looks like our terminal, duplicating stdin\n");
	    if (!ignorefds[fd_entry.fd]) {
		fd2 = syscall_check(dup2(stdinfd, fd_entry.fd), 0, "dup2");
		if (fd_entry.flags & FD_TERMIOS) {
		    extern int ioctl(int fd, unsigned long req, ...);
		    ioctl(fd_entry.fd, TCSETS, &fd_entry.termios);
		}
		//pid_t mypid = getpid();
		//ioctl(fd_entry.fd, TIOCSPGRP, &mypid);
	    }
	} else {
	    if (!ignorefds[fd_entry.fd]) {
		fd2 = open(fd_entry.filename, fd_entry.mode);
		if (fd2 < 0) {
		    fprintf(stderr, "Warning: couldn't restore file %s: %s\n", fd_entry.filename, strerror(errno));
		    continue;
		}
		if (!(fd_entry.flags & FD_OFFSET_NOT_SAVED)) {
		    if (verbosity > 1)
			fprintf(stderr, "    seeking to %ld\n", fd_entry.position);
		    if (lseek(fd2, fd_entry.position, SEEK_SET) < 0) {
			fprintf(stderr, "Warning: restoring file offset %ld to file %s failed: %s\n", fd_entry.position, fd_entry.filename, strerror(errno));
		    }
		}
		syscall_check(dup2(fd2, fd_entry.fd), 0, "dup2");
		syscall_check(close(fd2), 0, "close");
		fd2 = fd_entry.fd;
	    }
	}
	if (fd_entry.data_length >= 0) {
	    fprintf(stderr, "Warning: restoring file contents not yet implemented (for %s)\n", fd_entry.filename);
	    while (fd_entry.data_length) {
		int r = 4096;
		char buf[4096];
		if (r > fd_entry.data_length)
		    r = fd_entry.data_length;
		safe_read(fd, buf, r, "junk");
		fd_entry.data_length -= r;
	    }
	}
	fcntl(fd_entry.fd, F_SETFD, fd_entry.fcntl_data.close_on_exec);
    }
    //close(stdinfd);

    safe_read(fd, dir, sizeof(dir), "working directory");
    if (verbosity > 0)
	fprintf(stderr, "Changing directory to %s\n", dir);
    syscall_check(chdir(dir), 0, "chdir %s", dir);

    if (verbosity > 0)
	fprintf(stderr, "Restoring signal handlers...\n");
    for (i = 1; i <= MAX_SIGS; i++) {
	struct k_sigaction oksa;
	safe_read(fd, &sa, sizeof(sa), "sigaction");
	if (i == SIGKILL || i == SIGSTOP) continue;
	if (i == SIGSEGV && tls_hack) {
	    sa.sa_hand = (__sighandler_t)segv_handler;
	    sa.sa_flags |= SA_SIGINFO;
	    sa.sa_flags &= ~SA_ONESHOT;
	}
	if (set_rt_sigaction(i, &sa, (i==SIGSEGV)?&oksa:NULL) == -1) {
	    fprintf(stderr, "Couldn't restore signal handler %d: %s\n",
		    i, strerror(i));
	}
	if (i == SIGSEGV) {
	    old_segvhandler = (void*)oksa.sa_hand;
	}
    }

    close(fd);

    syscall_check(
	    (int)mmap((void*)TRAMPOLINE_ADDR, PAGE_SIZE, PROT_READ|PROT_WRITE|PROT_EXEC,
	    MAP_FIXED|MAP_PRIVATE|MAP_ANONYMOUS, 0, 0), 0, "mmap");
    /* put eflags onto the process' stack so we can pop it off */
    user_data.regs.esp-=4;
    *(long*)user_data.regs.esp = user_data.regs.eflags;
    
    put_shell_code(user_data.regs, (void*)0x10000);

    if (verbosity > 0)
	fprintf(stderr, "Ready to go!\n");

    /*
    if (translate_pids) {
	extern int supervise_me(pid_t);
	munmap((void*)RESUMER_END, 0xffffffff-RESUMER_END);
	supervise_me(pid);
    }
    */

    if (do_pause)
	sleep(2);
    asm("jmp 0x10000");
}

void handle_chunk(struct cp_chunk *chunk) {
    switch (chunk->type) {
	case CP_CHUNK_MISC:
	    process_chunk_misc(&chunk->misc);
	    break;
	case CP_CHUNK_REGS:
	    process_chunk_regs(&chunk->regs);
	    break;
	case CP_CHUNK_I387_DATA:
	    process_chunk_i387_data(&chunk->i387_data);
	    break;
	case CP_CHUNK_TLS:
	    process_chunk_tls(&chunk->tls);
	    break;
	case CP_CHUNK_FD:
	    process_chunk_fd(&chunk->fd);
	    break;
	case CP_CHUNK_VMA:
	    process_chunk_vma(&chunk->vma);
	    break;
	case CP_CHUNK_SIGHAND:
	    process_chunk_sighand(&chunk->sighand);
	    break;
    }
}

void read_process() {
    void *fptr;
    struct cp_chunk *chunk;

    fptr = stream_ops->init(image_fd, O_RDONLY);
    if (!fptr)
	bail("Unable to initialize reader.");

    /* Read and process all chunks. */
    while (chunk = read_chunk(fptr)) {
	handle_chunk(chunk);
    }

    /* Cleanup the input file. */
    stream_ops->finish(fptr);
    close(image_fd);

    /* The trampoline code should now be magically loaded at 0x10000.
     * Jumping there will restore registers and continue execution.
     */

    if (do_pause)
	sleep(2);

    asm("jmp 0x10000");
}

void* find_top_of_stack() {
    unsigned int tmp;
    /* Return the top of the current stack page. */
    return (void*)(((long)&tmp + PAGE_SIZE - 1) & ~(PAGE_SIZE-1));
}

void seek_to_image(int fd) {
    Elf32_Ehdr e;
    Elf32_Shdr s;
    int i;
    char* strtab;

    syscall_check(lseek(fd, 0, SEEK_SET), 0, "lseek");
    safe_read(fd, &e, sizeof(e), "Elf32_Ehdr");
    if (e.e_shoff == 0) {
	fprintf(stderr, "No section header found in self! Bugger.\n");
	exit(1);
    }
    if (e.e_shentsize != sizeof(Elf32_Shdr)) {
	fprintf(stderr, "Section headers incorrect size. Bugger.\n");
	exit(1);
    }
    if (e.e_shstrndx == SHN_UNDEF) {
	fprintf(stderr, "String section missing. Bugger.\n");
	exit(1);
    }
    
    /* read the string table */
    syscall_check(lseek(fd, e.e_shoff+(e.e_shstrndx*e.e_shentsize), SEEK_SET), 0, "lseek");
    safe_read(fd, &s, sizeof(s), "string table section header");
    syscall_check(lseek(fd, s.sh_offset, SEEK_SET), 0, "lseek");
    strtab = malloc(s.sh_size);
    safe_read(fd, strtab, s.sh_size, "string table");

    for (i=0; i < e.e_shnum; i++) {
	long offset;

	syscall_check(
		lseek(fd, e.e_shoff+(i*e.e_shentsize), SEEK_SET), 0, "lseek");
	safe_read(fd, &s, sizeof(s), "Elf32_Shdr");
	if (s.sh_type != SHT_PROGBITS || s.sh_name == 0)
	    continue;

	/* We have potential data! Is it really ours? */
	if (memcmp(strtab+s.sh_name, "cryopid.image", 13) != 0)
	    continue;

	if (s.sh_info != IMAGE_VERSION) {
	    fprintf(stderr, "Incorrect image version found (%d)! Keeping on trying.\n", s.sh_info);
	    continue;
	}

	/* Woo! got it! */
	syscall_check(
		lseek(fd, s.sh_offset, SEEK_SET), 0, "lseek");

	safe_read(fd, &offset, 4, "offset");

	syscall_check(
		lseek(fd, offset, SEEK_SET), 0, "lseek");

	return;
    }
    fprintf(stderr, "Program image not found! Bugger.\n");
    exit(1);
}

int open_self() {
    int fd;
    if (verbosity > 0)
	fprintf(stderr, "Reading image...\n");
    fd = open("/proc/self/exe", O_RDONLY);
    if (fd == -1) {
	fprintf(stderr, "Couldn't open self: %s\n", strerror(errno));
	exit(1);
    }
    return fd;
}

void usage(char* argv0) {
    fprintf(stderr,
"Usage: %s [options]\n"
"\n"
"This is a saved image of process. To resume this process, you can simply run\n"
"this executable. Some options that may be of interest when restoring\n"
"this process:\n"
"\n"
"    -v      Be verbose while resuming.\n"
"    -i <fd> Do not restore the given file descriptor.\n"
"    -p      Pause between steps before resuming (for debugging)\n"
"    -P      Attempt to gain original PID by way of fork()'ing a lot\n"
"    -t      Use ptrace to translate PIDs in system calls (Experimental and\n"
"	    incomplete!)"
"\n"
"This image was created by CryoPID. http://cryopid.berlios.de/\n",
    argv0);
    exit(1);
}

void real_main(int argc, char** argv) __attribute__((noreturn));
void real_main(int argc, char** argv) {
    int fd = 42;
    /* See if we're being executed for the second time. If so, read arguments
     * from the file.
     */
    if (lseek(fd, 0, SEEK_SET) != -1) {
	safe_read(fd, &argc, sizeof(argc), "argc from cryopid.state");
	argv = (char**)malloc(sizeof(char*)*argc+1);
	argv[argc] = NULL;
	int i, len;
	for (i=0; i < argc; i++) {
	    safe_read(fd, &len, sizeof(len), "argv len from cryopid.state");
	    argv[i] = (char*)malloc(len);
	    safe_read(fd, argv[i], len, "new argv from cryopid.state");
	}
	close(fd);
	reforked = 1;
    } else {
	if (errno != EBADF) {
	    /* EBADF is the only error we should be expecting! */
	    fprintf(stderr, "Unexpected error on lseek. Aborting (%s).\n",
		    strerror(errno));
	    exit(1);
	}
    }

    /* Parse options */
    memset(ignorefds, 0, sizeof(ignorefds));
    while (1) {
	int option_index = 0;
	int c;
	static struct option long_options[] = {
	    {0, 0, 0, 0},
	};
	
	c = getopt_long(argc, argv, "vpPc:ti:",
		long_options, &option_index);
	if (c == -1)
	    break;
	switch(c) {
	    case 'v':
		verbosity++;
		break;
	    case 'p':
		do_pause = 1;
		break;
	    case 'P':
		want_pid = 1;
		break;
	    case 't':
		translate_pids = 1;
		break;
	    case 'i':
		if (atoi(optarg) >= 256) {
		    fprintf(stderr, "Ignored fd number too high! Not ignoring.\n");
		    exit(1);
		}
		ignorefds[atoi(optarg)] = 1;
		break;
	    case '?':
		/* invalid option */
		fprintf(stderr, "Unknown option on command line.\n");
		usage(argv[0]);
		break;
	}
    }

    if (argc - optind) {
	fprintf(stderr, "Extra arguments not expected (%s ...)!\n", argv[optind]);
	usage(argv[0]);
    }

    fd = open_self();
    seek_to_image(fd);
    read_process(fd);

    fprintf(stderr, "Something went wrong :(\n");
    exit(1);
}

int main(int argc, char**argv) {
    long amount_used;
    void *stack_ptr;
    void *top_of_old_stack;
    void *top_of_new_stack;
    long size_of_new_stack;
    
    int i;

    /* Take a copy of our argc/argv and environment below we blow them away */
    real_argc = argc;
    real_argv = (char**)malloc((sizeof(char*)*argc)+1);
    for(i=0; i < argc; i++)
	real_argv[i] = strdup(argv[i]);
    real_argv[i] = NULL;

    for(i = 0; environ[i]; i++); /* count environment variables */
    real_environ = malloc((sizeof(char*)*i)+1);
    for(i = 0; environ[i]; i++)
	*real_environ++ = strdup(environ[i]);
    *real_environ = NULL;
    environ = real_environ;

    /* Reposition the stack at top_of_old_stack */
    top_of_old_stack = find_top_of_stack();
    stack_ptr = &stack_ptr;

    amount_used = top_of_old_stack - stack_ptr;

    top_of_new_stack = (void*)0x00200000;
    size_of_new_stack = PAGE_SIZE;

    syscall_check( (int)
	mmap(top_of_new_stack - size_of_new_stack, size_of_new_stack,
	    PROT_READ|PROT_WRITE|PROT_EXEC,
	    MAP_ANONYMOUS|MAP_FIXED|MAP_GROWSDOWN|MAP_PRIVATE, -1, 0),
	0, "mmap(newstack)");
    memset(top_of_new_stack - size_of_new_stack, 0, size_of_new_stack);
    memcpy(top_of_new_stack - size_of_new_stack,
	    top_of_old_stack - size_of_new_stack, /* FIX ME */
	    size_of_new_stack);
    __asm__ ("addl %0, %%esp" : : "a"(top_of_new_stack - top_of_old_stack));

    /* unmap absolutely everything above us! */
    syscall_check(
	    munmap(top_of_new_stack,
		(top_of_old_stack - top_of_new_stack)),
		0, "munmap(stack)");
    
    /* Now hope for the best! */
    real_main(real_argc, real_argv);
    /* should never return */
    return 42;
}

/* vim:set ts=8 sw=4 noet: */
